# coding: utf-8

class FileFormatError(Exception):
    pass

from .ase_Atoms import load as load_ase_Atoms
from .cif import load as load_cif
from .pymatgen_Structure import load as load_pymatgen_Structure
from .table import load as load_table
from .system_model import load as load_system_model
from .poscar import load as load_poscar
from .atom_data import load as load_atom_data
from .atom_dump import load as load_atom_dump
from .spglib_cell import load as load_spglib_cell
from .phonopy_Atoms import load as load_phonopy_Atoms

__all__ = ['FileFormatError', 'load', 'load_ase_Atoms', 'load_pymatgen_Structure',
           'load_table', 'load_system_model', 'load_poscar', 'load_atom_data',
           'load_atom_dump', 'load_cif', 'load_spglib_cell', 'load_phonopy_Atoms']

def load(style, input, **kwargs):
    """
    Load a System.
    
    Parameters
    ----------
    style : str
        Indicates the format of the content to load as an atomman.System
    input : str, file-like object or object
        The content to load.
    kwargs
        Any extra keyword arguments to pass to the underlying load methods.
        
    Returns
    -------
    system : atomman.System
        The system object associated with the data model.
    """
    
    if style == 'system_model':
        return load_system_model(input, **kwargs)
    
    elif style == 'cif':
        return load_cif(input, **kwargs)
    
    elif style == 'atom_data':
        return load_atom_data(input, **kwargs)
    
    elif style == 'atom_dump':
        return load_atom_dump(input, **kwargs)
    
    elif style == 'table':
        return load_table(input, **kwargs)
    
    elif style == 'ase_Atoms':
        return load_ase_Atoms(input, **kwargs)
    
    elif style == 'phonopy_Atoms':
        return load_phonopy_Atoms(input, **kwargs)
    
    elif style == 'pymatgen_Structure':
        return load_pymatgen_Structure(input, **kwargs)
    
    elif style == 'poscar':
        return load_poscar(input, **kwargs)
    
    elif style == 'spglib_cell':
        return load_spglib_cell(input, **kwargs)
    
    else:
        raise ValueError('Unsupported style')


# coding: utf-8

# atomman imports
from . import IsotropicVolterraDislocation, Stroh

def solve_volterra_dislocation(C, burgers, axes=None, m=[1,0,0], n=[0,1,0],
                               tol=1e-8):
    """
    Wrapper function for generating VolterraDislocation children classes
    that provide linear elastic solutions for straight dislocations.
    
    Parameters
    ----------
    C : atomman.ElasticConstants
        The medium's elastic constants.  Must be isotropic.
    burgers : array-like object
        The dislocation's Cartesian Burgers vector.
    axes : array-like object, optional
        3x3 set of rotational axes for the system. If given, C and burgers
        will be transformed using axes.
    m : array-like object, optional
        The m unit vector for the solution.  m, n, and u (dislocation
        line) should be right-hand orthogonal.  Default value is [1,0,0]
        (x-axis).
    n : array-like object, optional
        The n unit vector for the solution.  m, n, and u (dislocation
        line) should be right-hand orthogonal.  Default value is [0,1,0]
        (y-axis). n is normal to the dislocation slip plane.
    tol : float
        Tolerance parameter used to check for compatibility of the other
        parameters.  Default value is 1e-8.
    iso_atol : float
        The absolute tolerance to use when evaluating 
    iso_rtol : float
    style : str or None, optional
        Indicates which dislocation solution style to use.  Accepted values are
        'isotropic', 'anisotropic', or None.  If None (default), will select
        isotropic or anisotropic based on the supplied elastic constants.

    Returns
    atomman.defect.VolterraDislocation
        The dislocation solution of the appropriate type.
    """

    try:
        return Stroh(C, burgers, axes=axes, m=m, n=n, tol=tol)
    except:
        return IsotropicVolterraDislocation(C, burgers, axes=axes, m=m, n=n,
                                            tol=tol)
                                            


# coding: utf-8
# Standard Python imports
import os
from copy import deepcopy

# http://www.numpy.org/
import numpy as np

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

from ..tools import axes_check, vect_angle, miller

class StackingFault(object):
    def __init__(self, system, a1vect=None, a2vect=None,
                 ucellbox=None, transform=None, cutboxvector='c',
                 faultposrel=None, faultposcart=None, vacuumwidth=None):
        """
        Initializes an object for generating generalized stacking fault
        atomic configurations.

        Parameters
        ----------
        system : atomman.System
            The atomic system to generate stacking fault configurations from.
            The system should be oriented such that the stacking fault plane
            is parallel to either the xy-plane, the xz-plane or the yz-plane.
        a1vect : array-like object, optional
            A slip vector within the slip plane.  Depending on if ucellbox and
            transform are given, this can be either a Miller crystal vector or
            a Cartesian vector relative to the supplied system.  If a1vect is
            not given and a2vect is, then a1vect is set to [0,0,0].
        a2vect : array-like object, optional
            A slip vector within the slip plane.  Depending on if ucellbox and
            transform are given, this can be either a Miller crystal vector or
            a Cartesian vector relative to the supplied system.  If a2vect is
            not given and a1vect is, then a2vect is set to [0,0,0].
        ucellbox : atomman.Box, optional
            If ucellbox is given, then the a1vect and a2vect slip vectors are
            taken as Miller crystal vectors relative to ucellbox.  If not
            given, then the slip vectors are taken as Cartesian vectors.
        transform : array-like object, optional
            A transformation tensor to apply to the a1vect and a2vect slip
            vectors.  This is needed if system is oriented differently than
            ucellbox, i.e. system is rotated.
        cutboxvector : str, optional
            Defines which of the three box vectors gets cut so that the slip
            plane can be added.  'a' indicates the slip plane is parallel to
            the yz-plane, 'b' the xz-plane and 'c' the xy-plane.  Note that 
            the specified cutboxvector is the only box vector allowed to have a
            component out of the slip plane.
        faultposrel : float, optional
            The position to place the slip plane within the system given as a
            relative coordinate along the out-of-plane direction.  faultposrel
            and faultposcart cannot both be given.  Default value is 0.5 if 
            faultposcart is also not given.
        faultposcart : float, optional
            The position to place the slip plane within the system given as a
            Cartesian coordinate along the out-of-plane direction.  faultposrel
            and faultposcart cannot both be given.
        vacuumwidth : float, optional
            If given, the out-of-plane dimension of system is increased by this
            amount to create a vacuum region.  Useful if the generated
            configurations are to be evaluated with DFT.
        """
        
        self.set(system, a1vect=a1vect, a2vect=a2vect, ucellbox=ucellbox,
                    transform=transform, cutboxvector=cutboxvector,
                    faultposrel=faultposrel, faultposcart=faultposcart,
                    vacuumwidth=vacuumwidth)
        

    def set(self, system, a1vect=None, a2vect=None, ucellbox=None, transform=None,
            cutboxvector='c', faultposrel=None, faultposcart=None,
            vacuumwidth=None):
        """
        Sets generation parameters associated with a stacking fault.

        Parameters
        ----------
        system : atomman.System
            The atomic system to generate stacking fault configurations from.
            The system should be oriented such that the stacking fault plane
            is parallel to either the xy-plane, the xz-plane or the yz-plane.
        a1vect : array-like object, optional
            A slip vector within the slip plane.  Depending on if ucellbox and
            transform are given, this can be either a Miller crystal vector or
            a Cartesian vector relative to the supplied system.  If a1vect is
            not given and a2vect is, then a1vect is set to [0,0,0].
        a2vect : array-like object, optional
            A slip vector within the slip plane.  Depending on if ucellbox and
            transform are given, this can be either a Miller crystal vector or
            a Cartesian vector relative to the supplied system.  If a2vect is
            not given and a1vect is, then a2vect is set to [0,0,0].
        ucellbox : atomman.Box, optional
            If ucellbox is given, then the a1vect and a2vect slip vectors are
            taken as Miller crystal vectors relative to ucellbox.  If ucellbox
            is a standard hexagonal cell, the slip vectors can alternatively be
            given as Miller-Bravais crystal vectors. If not given, then the
            slip vectors are taken as Cartesian vectors.
        transform : array-like object, optional
            A transformation tensor to apply to the a1vect and a2vect slip
            vectors.  This is needed if system is oriented differently than
            ucellbox, i.e. system is rotated.
        cutboxvector : str, optional
            Defines which of the three box vectors gets cut so that the slip
            plane can be added.  'a' indicates the slip plane is parallel to
            the yz-plane, 'b' the xz-plane and 'c' the xy-plane.  Note that 
            the specified cutboxvector is the only box vector allowed to have a
            component out of the slip plane.
        faultposrel : float, optional
            The position to place the slip plane within the system given as a
            relative coordinate along the out-of-plane direction.  faultposrel
            and faultposcart cannot both be given.  Default value is 0.5 if 
            faultposcart is also not given.
        faultposcart : float, optional
            The position to place the slip plane within the system given as a
            Cartesian coordinate along the out-of-plane direction.  faultposrel
            and faultposcart cannot both be given.
        """
        # Copy system for safe manipulation
        system = deepcopy(system)
        system.wrap()

        # Check fault pos values
        if faultposcart is None and faultposrel is None:
            faultposrel = 0.5 
        elif faultposcart is not None and faultposrel is not None:
            raise ValueError('faultposrel and faultposcart cannot both be given')

        # Set cutindex and faultarea based on cutboxvector
        if cutboxvector == 'a':
            if system.box.bvect[0] != 0.0 or system.box.cvect[0] != 0.0:
                raise ValueError("box bvect and cvect cannot have x component for cutboxvector='a'")
            cutindex = 0
            faultarea = np.linalg.norm(np.cross(system.box.bvect, system.box.cvect))
        
        elif cutboxvector == 'b':
            if system.box.avect[1] != 0.0 or system.box.cvect[1] != 0.0:
                raise ValueError("box avect and cvect cannot have y component for cutboxvector='b'")
            cutindex = 1
            faultarea = np.linalg.norm(np.cross(system.box.avect, system.box.cvect))

        elif cutboxvector == 'c':
            if system.box.avect[2] != 0.0 or system.box.bvect[2] != 0.0:
                raise ValueError("box avect and bvect cannot have z component for cutboxvector='c'")
            cutindex = 2
            faultarea = np.linalg.norm(np.cross(system.box.avect, system.box.bvect))

        else: 
            raise ValueError('Invalid cutboxvector')

        # Define out of plane unit vector 
        ovect = np.zeros(3)
        ovect[cutindex] = 1.0

        # Set system's pbc
        system.pbc = [True, True, True]
        system.pbc[cutindex] = False

        # Add vacuum to non-periodic direction
        if vacuumwidth is not None:
            if vacuumwidth < 0:
                raise ValueError('vacuumwidth must be positive')
            newvects = system.box.vects
            newvects[cutindex, cutindex] += vacuumwidth
            neworigin = system.box.origin - ovect * vacuumwidth / 2 
            system.box_set(vects=newvects, origin=neworigin)

        # Identify atoms above fault plane position
        if faultposcart is None:
            faultposcart = system.box.origin[cutindex] + system.box.vects[cutindex, cutindex] * faultposrel            
        abovefault = system.atoms.pos[:, cutindex] > (faultposcart)

        # Set default a1vect or a2vect if needed
        if a1vect is None and a2vect is None:
            raise ValueError('At least one of a1vect and a2vect must be given')
        elif a1vect is None:
            a1vect = np.zeros(3)
        elif a2vect is None:
            a2vect = np.zeros(3)

        # Convert crystal vects to Cartesian
        if ucellbox is not None:
            a1vectcart = miller.vector_crystal_to_cartesian(a1vect, ucellbox)
            a2vectcart = miller.vector_crystal_to_cartesian(a2vect, ucellbox)
        else:
            a1vectcart = a1vect
            a2vectcart = a2vect
        
        # Transform vects to match system
        if transform is not None:
            transform = axes_check(transform)
            a1vectcart = transform.dot(a1vectcart)
            a2vectcart = transform.dot(a2vectcart)

        # Check shift vectors values
        if not np.isclose(a1vectcart[cutindex], 0.0):
            raise ValueError('a1vect not in slip plane')
        if not np.isclose(a2vectcart[cutindex], 0.0):
            raise ValueError('a2vect not in slip plane')
        if not np.allclose(a1vectcart, np.zeros(3)) and not np.allclose(a2vectcart, np.zeros(3)):
            angle = vect_angle(a1vectcart, a2vectcart)
            if np.isclose(angle, 0.0) or np.isclose(angle, 180.0):
                raise ValueError('a1vect and a2vect cannot be parallel')
    
        # Save properties
        self.__system = system
        self.__a1vect = a1vect
        self.__a2vect = a2vect
        self.__a1vectcart = a1vectcart
        self.__a2vectcart = a2vectcart
        self.__cutboxvector = cutboxvector
        self.__cutindex = cutindex
        self.__faultposcart = faultposcart
        self.__abovefault = abovefault
        self.__faultarea = faultarea
        self.__ucellbox = ucellbox
        self.__transform = transform

    @property
    def system(self):
        """atomman.System : The unshifted configuration."""
        return self.__system

    @property
    def a1vect(self):
        """numpy.ndarray : One of the two shift vectors as given."""
        return self.__a1vect

    @property
    def a2vect(self):
        """numpy.ndarray : One of the two shift vectors as given."""
        return self.__a2vect

    @property
    def a1vectcart(self):
        """numpy.ndarray : One of the two shift vectors in Cartesian relative to system."""
        return self.__a1vectcart
    
    @property
    def a2vectcart(self):
        """numpy.ndarray : One of the two shift vectors in Cartesian relative to system."""
        return self.__a2vectcart

    @property
    def cutboxvector(self):
        """str : The box vector that is cut to add vacuum and slip plane."""
        return self.__cutboxvector

    @property
    def cutindex(self):
        """int : The Cartesian index for the cut direction."""
        return self.__cutindex

    @property
    def faultposcart(self):
        """float : The Cartesian position of the slip plane."""
        return self.__faultposcart

    @property
    def abovefault(self):
        """list : Indices of all atoms in System above the slip plane."""
        return self.__abovefault
    
    @property
    def faultarea(self):
        """float : The area of the slip plane."""
        return self.__faultarea

    @property
    def ucellbox(self):
        """atomman.Box or None : The supplied reference cell box."""
        return self.__ucellbox

    @property
    def transform(self):
        """numpy.ndarray or None : The supplied transformation matrix."""
        return self.__transform
    
    def fault(self, a1=None, a2=None, outofplane=None, faultshift=None, minimum_r=None):
        """
        Generates a fault configuration by displacing all atoms above the slip
        plane.

        Parameters
        ----------
        a1 : float, optional
            The fractional coordinate of a1vect to shift by. 
            Default value is 0.0.
        a2 : float, optional
            The fractional coordinate of a2vect to shift by. 
            Default value is 0.0. 
        outofplane : float, optional
            An out-of-plane shift, given in absolute units.
            Default value is 0.0.
        faultshift : array-like object, optional
            The full shifting vector to displace the atoms above the slip
            plane by.  Cannot be given with a1, a2, or outofplane.
        minimum_r : float, optional
            Specifies the minimum allowed interatomic spacing across the slip
            plane.  If any sets of atoms are closer than this value then the
            outofplane shift is increased.  Default value is None, which
            performs no adjustment.

        Returns
        -------
        atomman.System
            The atomic configuration with stacking fault shift
        """
        # Define out of plane unit vector 
        ovect = np.zeros(3)
        ovect[self.cutindex] = 1.0
        
        # Identify the two non-cut indices
        inindex = []
        for i in range(3):
            if i != self.cutindex:
                inindex.append(i)
        
        # Calculate faultshift
        if a1 is not None or a2 is not None or outofplane is not None:
            if faultshift is not None:
                raise ValueError('a1, a2, outofplane cannot be given with faultshift')
            if a1 is None:
                a1 = 0.0
            if a2 is None:
                a2 = 0.0
            if outofplane is None:
                outofplane = 0.0
            faultshift = a1 * self.a1vectcart + a2 * self.a2vectcart + outofplane * ovect
        
        # Set default faultshift
        elif faultshift is None:
            faultshift = np.array([0.0, 0.0, 0.0])

        # Shift atoms above the fault by faultshift
        sfsystem = deepcopy(self.system)
        sfsystem.atoms.pos[self.abovefault] += faultshift
        sfsystem.wrap()
        
        # Add additional outofplane shift if necessary
        if minimum_r is not None:
            # Get all atoms within minimum_r of fault position
            top_pos = sfsystem.atoms.pos[self.abovefault]
            top_pos = top_pos[top_pos[:, self.cutindex] <= self.faultposcart + minimum_r] 
            bot_pos = sfsystem.atoms.pos[~self.abovefault]
            bot_pos = bot_pos[bot_pos[:, self.cutindex] >= self.faultposcart - minimum_r]
            if top_pos.shape[0] > 0 and bot_pos.shape[0] > 0:
                dmag_min = minimum_r
                dvect_min = None
                
                for i in range(top_pos.shape[0]):
                    dvect = sfsystem.dvect(bot_pos, top_pos[i])
                    if dvect.shape == (3,):
                        dvect = dvect.reshape(1,3)
                    dmag = np.linalg.norm(dvect, axis=1)
                    i = np.argmin(dmag)
                    if dmag[i] < dmag_min:
                        dmag_min = dmag[i]
                        dvect_min = dvect[i]
                
                if dvect_min is not None:
                    
                    new = (minimum_r**2 - dvect_min[inindex[0]]**2 - dvect_min[inindex[1]]**2)**0.5
                    outofplane = new - dvect_min[self.cutindex]
                    faultshift = outofplane * ovect
                    sfsystem.atoms.pos[self.abovefault] += faultshift
                    sfsystem.wrap()
        
        return sfsystem
    
    def iterfaultmap(self, num_a1=None, num_a2=None, outofplane=None, minimum_r=None):
        """
        Iterates over generalized stacking fault configurations associated
        with a 2D map of equally spaced a1, a2 coordinates.

        Parameters
        ----------
        num_a1 : int
            The number of a1 values to generate systems for.  
            Default value is 1 (only generate for a1=0.0).
        num_a2 : int
            The number of a2 values to generate systems for.  
            Default value is 1 (only generate for a2=0.0).
        outofplane : float, optional
            An out-of-plane shift, given in absolute units.
            Default value is 0.0.
        minimum_r : float, optional
            Specifies the minimum allowed interatomic spacing across the slip
            plane.  If any sets of atoms are closer than this value then the
            outofplane shift is increased.  Default value is None, which
            performs no adjustment.
        
        Yields
        ------
        a1 : float
            The a1 fractional coordinate of a1vect. 
        a2 : float
            The a2 fractional coordinate of a2vect. 
        atomman.System
            The fault configuration associated with the a1, a2 shift.
        """
        if num_a1 is None:
            num_a1 = 1
        if num_a2 is None:
            num_a2 = 1

        # Construct mesh of regular points
        a1s, a2s = np.meshgrid(np.linspace(0, 1, num_a1, endpoint=False),
                               np.linspace(0, 1, num_a2, endpoint=False))

        for a1, a2 in zip(a1s.flat, a2s.flat):
            yield a1, a2, self.fault(a1=a1, a2=a2, outofplane=outofplane,
                                     minimum_r=minimum_r)


# coding: utf-8
# Standard Python libraries
import os

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# atomman imports
import atomman.unitconvert as uc
from ..tools import atomic_mass, aslist, uber_open_rmode

class Potential(object):
    """
    Class for building LAMMPS input lines from a potential-LAMMPS data model.
    """
    
    def __init__(self, model, pot_dir=None):
        """
        Initializes an instance associated with a potential-LAMMPS data model.
        
        Parameters
        ----------
        model : str or file-like object
            A JSON/XML data model containing a potential-LAMMPS branch.
        pot_dir : str, optional
            The path to a directory containing any artifacts associated with
            the potential.  Default value is None, which assumes any required
            files will be in the working directory when LAMMPS is executed.
        """
        self.load(model, pot_dir)
    
    def load(self, model, pot_dir=None):
        """
        Loads potential-LAMMPS data model info.
        
        Parameters
        ----------
        model : str or file-like object
            A JSON/XML data model containing a potential-LAMMPS branch.
        pot_dir : str, optional
            The path to a directory containing any artifacts associated with
            the potential.  Default value is None, which assumes any required
            files will be in the working directory when LAMMPS is executed.
        """
        
        # Load model and find potential-LAMMPS
        self.__dm = DM(model).find('potential-LAMMPS')
        
        # Extract properties
        self.__id = self.__dm['id']
        self.__key = self.__dm['key']
        self.__potid = self.__dm['potential']['id']
        self.__potkey = self.__dm['potential']['key']
        self.__units = self.__dm['units']
        self.__atom_style = self.__dm['atom_style']
        self.__pair_style = self.__dm['pair_style']['type']

        allsymbols = self.__dm.get('allsymbols', 'False')
        if allsymbols is True or allsymbols == 'True':
            self.__allsymbols = True
        elif allsymbols is False or allsymbols == 'False':
            self.__allsymbols = False
        else:
            raise ValueError('Invalid allsymbols value')

        self.__status = self.__dm.get('status', 'active')
        
        if pot_dir is not None:
            self.pot_dir = pot_dir
        else:
            self.pot_dir = ''
        
        # Build lists of symbols, elements and masses
        self.__elements = []
        self.__symbols = []
        self.__masses = []
        self.__charges = []
        for atom in self.__dm.iteraslist('atom'):
            element = atom.get('element', None)
            symbol = atom.get('symbol', None)
            mass = atom.get('mass', None)
            charge = float(atom.get('charge', 0.0))
            
            # Check if element is listed
            if element is None:
                if mass is None:
                    raise KeyError('mass is required for each atom if element is not listed')
                if symbol is None:
                    raise KeyError('symbol is required for each atom if element is not listed')
                else:
                    element = symbol
            
            # Check if symbol is listed.
            if symbol is None:
                symbol = element
            
            # Check if mass is listed.
            if mass is None:
                mass = atomic_mass(element)
            else:
                mass = float(mass)
            
            # Add values to the lists
            self.__elements.append(element)
            self.__symbols.append(symbol)
            self.__masses.append(mass)
            self.__charges.append(charge)
    
    def normalize_symbols(self, symbols):
        """
        Modifies a given list of symbols to be compatible with the potential.
        Mostly, this converts symbols to a list if it is not already one, and
        adds additional symbols if the potential's allsymbols setting is True.

        Parameters
        ----------
        symbols : str or list-like object
            The initial list of symbols
        
        Returns
        -------
        list
            The updated list.
        """
        
        # Convert symbols to a list if needed
        symbols = aslist(symbols)
        
        # Check that all symbols are set
        for symbol in symbols:
            assert symbol is not None, 'symbols list incomplete: found None value'
        
        # Add missing symbols if potential's allsymbols is True
        if self.allsymbols:
            for symbol in self.symbols:
                if symbol not in symbols:
                    symbols.append(symbol)
        
        return symbols

    def __str__(self):
        """str: The string of the Potential returns its human-readable id"""
        return self.id
    
    @property
    def pot_dir(self):
        """str : The directory containing files associated with a given potential."""
        return self.__pot_dir
    
    @pot_dir.setter
    def pot_dir(self, value):
        self.__pot_dir = str(value)
    
    @property
    def id(self):
        """str : Human-readable identifier for the LAMMPS implementation."""
        return self.__id
    
    @property
    def key(self):
        """str : uuid hash-key for the LAMMPS implementation."""
        return self.__key
    
    @property
    def potid(self):
        """str : Human-readable identifier for the potential model."""
        return self.__potid
    
    @property
    def potkey(self):
        """str : uuid hash-key for the potential model."""
        return self.__potkey
    
    @property
    def units(self):
        """str : LAMMPS units option."""
        return self.__units
    
    @property
    def atom_style(self):
        """str : LAMMPS atom_style option."""
        return self.__atom_style
    
    @property
    def symbols(self):
        """list of str : All atom-model symbols."""
        return self.__symbols
    
    @property
    def pair_style(self):
        return self.__pair_style
    
    @property
    def allsymbols(self):
        """bool : Indicates if all atom-model symbols must be listed by pair_info."""
        return self.__allsymbols

    @property
    def status(self):
        """str : Indicates the status of the implementation (active, superseded, retracted)"""
        return self.__status

    def elements(self, symbols=None):
        """
        Returns a list of element names associated with atom-model symbols.
        
        Parameters
        ----------
        symbols : list of str, optional
            A list of atom-model symbols.  If None (default), will use all of
            the potential's symbols.
        
        Returns
        -------
        list of str
            The str element symbols corresponding to the atom-model symbols.
        """
        # Return all elements if symbols is None
        if symbols is None:
            return self.__elements
        
        # Normalize symbols
        symbols = self.normalize_symbols(symbols)
        
        # Get all matching elements
        elements = []
        for symbol in symbols:
            i = self.symbols.index(symbol)
            elements.append(self.__elements[i])
        
        return elements
        
    def masses(self, symbols=None):
        """
        Returns a list of atomic/ionic masses associated with atom-model
        symbols.
        
        Parameters
        ----------
        symbols : list of str, optional
            A list of atom-model symbols.  If None (default), will use all of
            the potential's symbols.
        
        Returns
        -------
        list of float
            The atomic/ionic masses corresponding to the atom-model symbols.
        """
        # Return all masses if symbols is None
        if symbols is None:
            return self.__masses
        
        # Normalize symbols
        symbols = self.normalize_symbols(symbols)
        
        # Get all matching masses
        masses = []
        for symbol in symbols:
            i = self.symbols.index(symbol)
            masses.append(self.__masses[i])
        
        return masses

    def charges(self, symbols=None):
        """
        Returns a list of atomic charges associated with atom-model symbols.
        Will have a None value if not assigned.
        
        Parameters
        ----------
        symbols : list of str, optional
            A list of atom-model symbols.  If None (default), will use all of
            the potential's symbols.
        
        Returns
        -------
        list of float
            The atomic charges corresponding to the atom-model symbols.
        """
        # Return all charges if symbols is None
        if symbols is None:
            return self.__charges
        
        # Normalize symbols
        symbols = self.normalize_symbols(symbols)
        
        # Get all matching charges
        charges = []
        for symbol in symbols:
            i = self.symbols.index(symbol)
            charges.append(self.__charges[i])
        
        return charges

    def pair_info(self, symbols=None, masses=None):
        """
        Generates the LAMMPS input command lines associated with the Potential
        and a list of atom-model symbols.
        
        Parameters
        ----------
        symbols : list of str, optional
            List of atom-model symbols corresponding to the atom types in a
            system.  If None (default), then all atom-model symbols will
            be included in the order that they are listed in the data model.
        masses : list, optional
            Can be given to override the default symbol-based masses for each
            atom type.  Must be a list of the same length as symbols.  Any
            values of None in the list indicate that the default value be used
            for that atom type.
        
        Returns
        -------
        str
            The LAMMPS input command lines that specifies the potential.
        """
        # If no symbols supplied use all for potential
        if symbols is None:
            symbols = self.symbols
        else:
            symbols = aslist(symbols)

        # Check length of masses
        if masses is not None:
            masses = aslist(masses)
            assert len(masses) == len(symbols), 'supplied masses must be same length as symbols'
        
        # Normalize symbols
        symbols = self.normalize_symbols(symbols)
        
        # Set masses
        defaultmasses = self.masses(symbols)
        if masses is not None:
            for i in range(len(masses)):
                if masses[i] is not None:
                    defaultmasses[i] = masses[i]
        masses = defaultmasses
        
        info = ''

        # Generate mass lines
        for i in range(len(masses)):
            info += f'mass {i+1} {masses[i]}\n'
        info +='\n'
        
        # Generate pair_style line
        terms = self.__dm['pair_style'].aslist('term')
        style_terms = self.__pair_terms(terms)
        
        info += f'pair_style {self.pair_style}{style_terms}\n'
        
        # Generate pair_coeff lines
        for coeff_line in self.__dm.iteraslist('pair_coeff'):
            if 'interaction' in coeff_line:
                interaction = coeff_line['interaction'].get('symbol', ['*', '*'])
            else:
                interaction = ['*', '*']
            
            # Always include coeff lines that act on all atoms in the system
            if interaction == ['*', '*']:
                coeff_symbols = self.symbols
                coeff_terms = self.__pair_terms(coeff_line.iteraslist('term'),
                                                symbols, coeff_symbols)
                
                info += f'pair_coeff * *{coeff_terms}\n'
                continue
            
            # Many-body potentials will contain a symbols term
            if len(coeff_line.finds('symbols')) > 0:
                many = True
            else:
                many = False
            
            # Treat as a many-body potential
            if many:
                coeff_symbols = interaction
                coeff_terms = self.__pair_terms(coeff_line.iteraslist('term'),
                                           symbols, coeff_symbols) + '\n'
                
                info += f'pair_coeff * *{coeff_terms}\n' 
            
            # Treat as pair potential
            else:
                coeff_symbols = interaction
                if len(coeff_symbols) != 2:
                    raise ValueError('Pair potential interactions need two listed elements')
                
                # Classic eam style is a special case
                if self.pair_style == 'eam':
                    if coeff_symbols[0] != coeff_symbols[1]:
                        raise ValueError('Only i==j interactions allowed for eam style')
                    for i in range(len(symbols)):
                        if symbols[i] == coeff_symbols[0]:
                            coeff_terms = self.__pair_terms(coeff_line.iteraslist('term'),
                                                           symbols, coeff_symbols)
                            
                            info += f'pair_coeff {i+1} {i+1}{coeff_terms}\n'
                
                # All other pair potentials
                else:
                    for i in range(len(symbols)):
                        for j in range(i, len(symbols)):
                            if ((symbols[i] == coeff_symbols[0] and symbols[j] == coeff_symbols[1])
                             or (symbols[i] == coeff_symbols[1] and symbols[j] == coeff_symbols[0])):
                                coeff_terms = self.__pair_terms(coeff_line.iteraslist('term'),
                                                                symbols, coeff_symbols)
                                
                                info += f'pair_coeff {i+1} {j+1}{coeff_terms}\n'
        
        # Generate additional command lines
        for command_line in self.__dm.iteraslist('command'):
            info += self.__pair_terms(command_line.iteraslist('term'),
                                         symbols, self.symbols).strip() + '\n'
        
        return info
    
    def __pair_terms(self, terms, system_symbols = [], coeff_symbols = []):
        """utility function used by self.pair_info() for composing lines from terms"""
        line = ''
        
        for term in terms:
            for ttype, tval in term.items():
                # Print options and parameters as strings
                if ttype == 'option' or ttype == 'parameter':
                    line += ' ' + str(tval)
                
                # Print files with pot_dir prefixes
                elif ttype == 'file':
                    line += ' ' + str(os.path.join(self.pot_dir, tval))
                
                # Print all symbols being used for symbolsList
                elif ttype == 'symbolsList' and (tval is True or tval == 'True'):
                    for coeff_symbol in coeff_symbols:
                        if coeff_symbol in system_symbols:
                            line += ' ' + coeff_symbol
                
                # Print symbols being used with model in appropriate order for symbols
                elif ttype == 'symbols' and (tval is True or tval == 'True'):
                    for system_symbol in system_symbols:
                        if system_symbol in coeff_symbols:
                            line += ' ' + system_symbol
                        else:
                            line += ' NULL'
        
        return line


# coding: utf-8
# Standard Python libraries
import os
from copy import deepcopy
import warnings

# http://www.numpy.org/
import numpy as np

# https://www.scipy.org/
from scipy.optimize import minimize

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# atomman imports
import atomman.unitconvert as uc
from ..tools import axes_check
from .GammaSurface import GammaSurface

class SDVPN(object):
    """
    Class representation of the semidiscrete variational Peierls-Nabarro
    dislocation model.
    """
    
    def __init__(self, *args, **kwargs):
        """
        Class initializer. Calls either load() or solve() based on the given
        parameters.
        
        Parameters
        ----------
        *args : list 
            Any unnamed arguments passed on to set().
        *kwargs : dict 
            Any named arguments to pass on.  If 'model' is in kwargs, then
            model() is called.  Else, *args and **kwargs are saved as class
            properties, and then solve() is called.
        """
        
        # Load solution from existing model
        if 'model' in kwargs:
            gamma = kwargs.pop('gamma', None)
            if len(kwargs) == 1 and len(args) == 0:
                self.load(kwargs['model'], gamma=gamma)
            else:
                raise ValueError('model keyword cannot be given with calculation parameters')
        
        # Solve new problem
        else:
            # Set default parameters
            self.__set(*args, **kwargs)
            
            # Call solve
            self.solve()
    
    def __set(self, x, disregistry, gamma, axes, K_tensor, **kwargs):
        """
        Sets default values for class parameters.
        
        Parameters
        ----------
        x : numpy.ndarray
            An array of shape (N) giving the x coordinates corresponding to
            the disregistry solution.
        disregistry : numpy.ndarray
            A (N,3) array giving the initial disregistry vector guess at each
            x coordinate.
        gamma : atomman.defect.GammaSurface
            The gamma surface (stacking fault map) to use for computing the
            misfit energy.
        axes : numpy.ndarray
            A (3,3) array defining the crystal directions of the dislocation
            system. Used to orient the disregistry vector to the gamma surface
            in computing the misfit energy.
        K_tensor : numpy.ndarray
            A (3,3) array giving the anisotropic elastic energy coefficients
            corresponding to the dislocation system's orientation.  Can be
            computed with atomman.defect.Stroh.
        tau : numpy.ndarray, optional
            A (3,3) array giving the stress tensor to apply to the system
            using the stress energy term.  Only the xy, yy, and yz components
            are used.  Default value is all zeros.
        alpha : list of float, optional
            The alpha coefficient(s) used by the nonlocal energy term.  Default
            value is [0.0].
        beta : numpy.ndarray, optional
            The (3,3) array of beta coefficient(s) used by the surface energy
            term.  Default value is all zeros.
        cutofflongrange : float, optional
            The cutoff distance to use for computing the long-range energy.
            Default value is 1000 angstroms.
        burgers : numpy.ndarray, optional
            The (3,) array of the dislocation's Burgers vector relative to the
            dislocation system.  Used only by the long-range energy.  Default
            value is all zeros (long-range energy will be excluded).
        fullstress : bool, optional
            Flag indicating which stress energy algorithm to use.  Default
            value is True.
        cdiffelastic : bool, optional
            Flag indicating if the dislocation density for the elastic energy
            component is computed with central difference (True) or simply
            neighboring values (False).  Default value is False.
        cdiffsurface : bool, optional
            Flag indicating if the dislocation density for the surface energy
            component is computed with central difference (True) or simply
            neighboring values (False).  Default value is True.
        cdiffstress : bool, optional
            Flag indicating if the dislocation density for the stress energy
            component is computed with central difference (True) or simply
            neighboring values (False).  Only matters if fullstress is True.
            Default value is False.
        min_method : str, optional
            The scipy.optimize.minimize method to use.  Default value is
            'Powell'.
        min_options : dict, optional
            Any options to pass on to scipy.optimize.minimize. Default value
            is {}.
        """
        
        # Set mandatory parameters
        self.__x = x
        self.__disregistry = disregistry
        self.__gamma = gamma
        self.__axes = axes
        self.__K_tensor = K_tensor
        
        # Set optional keyword parameters
        self.__tau = kwargs.get('tau', np.zeros((3,3)))
        self.__alpha = kwargs.get('alpha', [0.0])
        if not isinstance(self.__alpha, list):
            self.__alpha = [self.__alpha]
        self.__beta = kwargs.get('beta', np.zeros((3,3)))
        self.__cutofflongrange = kwargs.get('cutofflongrange',
                                            uc.set_in_units(1000, 'angstrom'))
        self.__burgers = kwargs.get('burgers', np.zeros(3))
        self.__fullstress = kwargs.get('fullstress', True)
        self.__cdiffelastic = kwargs.get('cdiffelastic', False)
        self.__cdiffsurface = kwargs.get('cdiffsurface', True)
        self.__cdiffstress = kwargs.get('cdiffstress', False)
        self.__min_method = kwargs.get('min_method', 'Powell')
        self.__min_options = kwargs.get('min_options', {})
    
    def solve(self, **kwargs):
        """
        Solves the semidiscrete variational Peierls-Nabarro dislocation
        disregistry through energy minimization using the set class
        properties.  All parameters are optional keyword arguments that
        replace the class properties.
        
        Parameters
        ----------
        x : numpy.ndarray, optional
            An array of shape (N) giving the x coordinates corresponding to
            the disregistry solution.
        disregistry : numpy.ndarray, optional
            A (N,3) array giving the initial disregistry vector guess at each
            x coordinate.
        gamma : atomman.defect.GammaSurface, optional
            The gamma surface (stacking fault map) to use for computing the
            misfit energy.
        axes : numpy.ndarray, optional
            A (3,3) array defining the crystal directions of the dislocation
            system. Used to orient the disregistry vector to the gamma surface
            in computing the misfit energy.
        K_tensor : numpy.ndarray, optional
            A (3,3) array giving the anisotropic elastic energy coefficients
            corresponding to the dislocation system's orientation.  Can be
            computed with atomman.defect.Stroh.
        tau : numpy.ndarray, optional
            A (3,3) array giving the stress tensor to apply to the system
            using the stress energy term.  Only the xy, yy, and yz components
            are used.
        alpha : list of float, optional
            The alpha coefficient(s) used by the nonlocal energy term.
        beta : numpy.ndarray, optional
            The (3,3) array of beta coefficient(s) used by the surface energy
            term.
        cutofflongrange : float, optional
            The cutoff distance to use for computing the long-range energy.
        burgers : numpy.ndarray, optional
            The (3,) array of the dislocation's Burgers vector relative to the
            dislocation system.  Used only by the long-range energy.
        fullstress : bool, optional
            Flag indicating which stress energy algorithm to use.
        cdiffelastic : bool, optional
            Flag indicating if the dislocation density for the elastic energy
            component is computed with central difference (True) or simply
            neighboring values (False).
        cdiffsurface : bool, optional
            Flag indicating if the dislocation density for the surface energy
            component is computed with central difference (True) or simply
            neighboring values (False).
        cdiffstress : bool, optional
            Flag indicating if the dislocation density for the stress energy
            component is computed with central difference (True) or simply
            neighboring values (False).  Only matters if fullstress is True.
        min_method : str, optional
            The scipy.optimize.minimize method to use.
        min_options : dict, optional
            Any options to pass on to scipy.optimize.minimize.
        """
        
        # Define subfunctions
        def decompose(d):
            """Breaks disregistry into components for minimization"""
            d13 = np.concatenate([d[1:-1, 0], d[1:-1, 2]])
            first = d[0]
            last = d[-1]
            return d13, first, last

        def recompose(d13, first, last):
            """Reassembles the disregistry components"""
            half = int(len(d13)/2)
            d = np.zeros((half+2, 3))
            d[0] = first
            d[-1] = last
            d[1:-1, 0] = d13[:half]
            d[1:-1, 2] = d13[half:]
            return d

        def min_func(d13, first, last):
            """Function for minimizing"""
            disregistry = recompose(d13, first, last)
            return self.total_energy(disregistry=disregistry)
        
        # Handle parameters
        self.__x = kwargs.pop('x', self.__x)
        self.__disregistry = kwargs.pop('disregistry', self.__disregistry)
        self.__gamma = kwargs.pop('gamma', self.__gamma)
        self.__axes = kwargs.pop('axes', self.__axes)
        self.__K_tensor = kwargs.pop('K_tensor', self.__K_tensor)
        self.__tau = kwargs.pop('tau', self.__tau)
        self.__alpha = kwargs.pop('alpha', self.__alpha)
        if not isinstance(self.__alpha, list):
            self.__alpha = [self.__alpha]
        self.__beta = kwargs.pop('beta', self.__beta)
        self.__cutofflongrange = kwargs.pop('cutofflongrange', 
                                            self.__cutofflongrange)
        self.__burgers = kwargs.pop('burgers', self.__burgers)
        self.__fullstress = kwargs.pop('fullstress', self.__fullstress)
        self.__cdiffelastic = kwargs.pop('cdiffelastic', self.__cdiffelastic)
        self.__cdiffsurface = kwargs.pop('cdiffsurface', self.__cdiffsurface)
        self.__cdiffstress = kwargs.pop('cdiffstress', self.__cdiffstress)
        self.__min_method = kwargs.pop('min_method', self.__min_method)
        self.__min_options = kwargs.pop('min_options', self.__min_options)
        
        # Solve disregistry
        d13, first, last = decompose(self.disregistry)
        res = minimize(min_func, d13, args=(first, last),
                       method=self.min_method, options=self.min_options)
        self.__disregistry = recompose(res.x, first, last)
        
        self.__res = res
    
    @property
    def x(self):
        """numpy.ndarray : The x coordinates."""
        return deepcopy(self.__x)
    
    @property
    def disregistry(self):
        """numpy.ndarray : The disregistry vector for each x coordinate."""
        return deepcopy(self.__disregistry)
    
    @property
    def gamma(self):
        """atomman.defect.GammaSurface : The stacking fault map."""
        return deepcopy(self.__gamma)
    
    @property
    def axes(self):
        """numpy.ndarray : Axes for orienting the disregistry to gamma."""
        return deepcopy(self.__axes)
    
    @property
    def K_tensor(self):
        """numpy.ndarray : Dislocation energy coefficient tensor for system."""
        return deepcopy(self.__K_tensor)
    
    @property
    def tau(self):
        """numpy.ndarray : The applied stress state."""
        return deepcopy(self.__tau)
    
    @property
    def alpha(self):
        """list of float : Coefficients for nonlocal energy correction."""
        return deepcopy(self.__alpha)
    
    @property
    def beta(self):
        """numpy.ndarray : Coefficients for gradient energy correction."""
        return deepcopy(self.__beta)
    
    @property
    def cutofflongrange(self):
        """float : Cutoff distance for long-range elastic energy."""
        return self.__cutofflongrange
    
    @property
    def burgers(self):
        """numpy.ndarray : Full Burgers vector for long-range elastic energy."""
        return deepcopy(self.__burgers)
    
    @property
    def fullstress(self):
        """bool : Flag indicating which stress algorithm was used."""
        return self.__fullstress
    
    @property
    def cdiffelastic(self):
        """bool : Flag indicating if elastic energy used central difference for computing the dislocation density."""
        return self.__cdiffelastic
    
    @property
    def cdiffsurface(self):
        """bool : Flag indicating if surface energy used central difference for computing the dislocation density."""
        return self.__cdiffsurface
    
    @property
    def cdiffstress(self):
        """bool : Flag indicating if stress energy used central difference for computing the dislocation density."""
        return self.__cdiffstress
    
    @property
    def min_method(self):
        """str : scipy.optimize.minimize method used."""
        return self.__min_method
    
    @property
    def min_options(self):
        """dict : scipy.optimize.minimize options used."""
        return self.__min_options
        
    @property
    def res(self):
        """OptimizeResult : scipy.optimize.minimize result."""
        return self.__res
    
    def disldensity(self, x=None, disregistry=None, cdiff=False):
        """
        Computes the dislocation density as the numerical derivative of
        disregistry with respect to x.  Uses either neighboring values
        
            ρ[i] = (δ[i] - δ[i-1]) / (x[i] - x[i-1])
        
        or central difference
        
            ρ[i] = (δ[i+1] - δ[i-1]) / (x[i+1] - x[i-1])
        
        Parameters
        ----------
        cdiff : bool
            Flag indicating how to compute the derivative.  A value of False
            (default) compares nearest values while a value of True compares
            next-nearest (i.e., central difference).
        
        Returns
        -------
        newx : numpy.array
            The x positions corresponding to the dislocation density values.
        rho : numpy.array
            The computed dislocation density.
        """
        # Default values are class properties
        if x is None:
            x = self.__x
        if disregistry is None:
            disregistry = self.__disregistry
        
        # Extract values
        d = disregistry
        
        if cdiff is False:
            # ρ[i] = (δ[i] - δ[i-1]) / (x[i] - x[i-1])
            rho = ((d[1:] - d[:-1]).T / (x[1:] - x[:-1])).T
            
            # newx is all x except the first
            newx = x[1:]
        
        elif cdiff is True:
            # ρ[i] = (δ[i+1] - δ[i-1]) / (x[i+1] - x[i-1])
            rho = ((d[2:] - d[:-2]).T / (x[2:] - x[:-2])).T
            
            # newx is all x except the first and last
            newx = x[1:-1]
        else:
            raise TypeError('cdiff must be bool')
            
        return (newx, rho)
    
    def misfit_energy(self, x=None, disregistry=None):
        """
        Computes the misfit energy for the disregistry using the stored gamma
        surface
        
            E_misfit = Σ γ(δ)Δx
        
        Parameters
        ----------
        x : numpy.ndarray, optional
            x-coordinates.  Default value is the stored x-coordinates.
        disregistry : numpy.ndarray, optional
            (N, 3) shaped array of disregistry vectors at each x-coordinate.
            Default value is the stored disregistry values.
            
        Returns
        -------
        float
            The misfit energy for the dislocation.
        """
        # Default values are class properties
        if x is None:
            x = self.__x
        if disregistry is None:
            disregistry = self.__disregistry
        
        # Extract values
        d = disregistry
        dx = x[1] - x[0]
        axes = self.__axes
        gamma = self.__gamma
        
        # Strip out y-component of disregistry and transform for gamma
        disreg = np.vstack([d[:,0], np.zeros(len(d)), d[:,2]]).T
        T = axes_check(axes)
        pos = np.inner(disreg, T.T)
        
        # Σ γ(δ)Δx
        return dx * gamma.E_gsf(pos=pos).sum()
    
    def elastic_energy(self, x=None, disregistry=None):
        r"""
        Computes the short-range configuration-dependent elastic energy term
        for the dislocation based on the dislocation density and K_tensor.
        
            E_elastic = 1/(4π) Σ_i Σ_j χ(i,j,Δx) K_lm ρ_l[i] ρ_m[j]
            
            χ(i,j,Δx) = (3/2) Δx² + ψ(i-1,j-1,Δx) + ψ(i,j,Δx) - ψ(i,j-1,Δx) - ψ(j,i-1,Δx)
            
            ψ(i,j,Δx) = (1/2) (i-j)² Δx² ln(\|i-j\|Δx)
        
        Parameters
        ----------
        x : numpy.ndarray, optional
            x-coordinates.  Default value is the stored x-coordinates.
        disregistry : numpy.ndarray, optional
            (N, 3) shaped array of disregistry vectors at each x-coordinate.
            Default value is the stored disregistry values.
        
        Returns
        -------
        float
            The elastic energy for the dislocation.
        """
        # Define subfunctions
        def chi(i, j, dx):
            """
            Computes the chi subfunction:
                χ(i,j,Δx) = (3/2) Δx² + ψ(i-1,j-1,Δx) + ψ(i,j,Δx)
                                      - ψ(i,j-1,Δx) - ψ(j,i-1,Δx)
            """
            return 3./2. * dx**2 + (psi(i-1, j-1, dx) + psi(i, j, dx)
                                    - psi(i, j-1, dx) - psi(j, i-1, dx))

        def psi(i, j, dx):
            """
            Computes the psi subfunction:
                ψ(i,j,Δx) = (1/2) (i-j)² Δx² ln(|i-j|Δx)
            """
            # Suppress NaN runtime warnings
            with warnings.catch_warnings():
                warnings.simplefilter("ignore", category=RuntimeWarning)
                
                # ψ(i,j,Δx) = (1/2) (i-j)² Δx² ln(|i-j|Δx)
                p = 0.5 * (i - j)**2 * dx**2 * np.log(np.abs(i - j) * dx)
            
            # Replace NaN values with 0.0
            p[np.isnan(p)] = 0.0
            
            return p
        
        # Default values are class properties
        if x is None:
            x = self.__x
        if disregistry is None:
            disregistry = self.__disregistry
        
        # Extract values
        d = disregistry
        dx = x[1] - x[0]
        cdiff = self.cdiffelastic
        Kij = self.__K_tensor
        
        rho = self.disldensity(x=x, disregistry=d, cdiff=cdiff)[1]
        j = np.arange(len(rho), dtype=int)
        
        energy = 0.0
        
        # Compute elastic energy (looping over i, vectorization over j)
        # 1/(4π) Σ_i Σ_j χ(i,j,Δx) K_lm ρ_l[i] ρ_m[j]
        for i in j:
            energy += np.sum(chi(i, j, dx) * np.inner(rho[i].dot(Kij), rho)) / (4 * np.pi)
        
        return energy
    
    def longrange_energy(self):
        """
        Computes the long-range elastic energy term for the dislocation using
        the K_tensor, Burgers vector and long-range cutoff.  This term is
        configuration-independent and thus the method takes no parameters.
        
            E_longrange = 1/(2π) K_lm b_l b_m ln(L)
        
        Returns
        -------
        float
            The long-range energy for the dislocation.
        """
        # Extract values
        Kij = self.__K_tensor
        b = self.__burgers
        L = self.__cutofflongrange
        
        # Compute long-range energy
        # 1/(2π) K_lm b_l b_m ln(L)
        return np.inner(b.dot(Kij), b) * np.log(L) / (2 * np.pi)
    
    def stress_energy(self, x=None, disregistry=None):
        """
        Computes the stress energy due to the applied stress, tau.
        If fullstress is True:
        
            E_stress = -1/2 Σ_i (x[i]² - x[i-1]²) ρ_l τ_2l
        
        If fullstress is False:
        
            E_stress = -1/2 Σ_i τ_2l (δ_l[i] + δ_l[i+1]) Δx
        
        Parameters
        ----------
        x : numpy.ndarray, optional
            x-coordinates.  Default value is the stored x-coordinates.
        disregistry : numpy.ndarray, optional
            (N, 3) shaped array of disregistry vectors at each x-coordinate.
            Default value is the stored disregistry values.
        
        Returns
        -------
        float
            The stress energy for the dislocation.
        """
        # Default values are class properties
        if x is None:
            x = self.__x
        if disregistry is None:
            disregistry = self.__disregistry
        
        # Extract values
        d = disregistry
        dx = x[1] - x[0]
        tau = self.__tau
        full = self.fullstress
        cdiff = self.cdiffstress
        
        if full is True:
            rho = self.disldensity(x=x, disregistry=d, cdiff=cdiff)[1]
            # -1/2 Σ_i (x[i]² - x[i-1]²) ρ_l τ_2l
            return -0.5 * np.sum((x[1:]**2 - x[:-1]**2) * np.inner(rho, tau[1,:]))
            
        else:
            # -1/2 Σ_i τ_2l (δ_l[i] + δ_l[i+1]) Δx
            return -0.5 * np.sum(np.inner(tau[1,:], (d[:-1] + d[1:]) * dx))
    
    def surface_energy(self, x=None, disregistry=None):
        """
        Computes the gradient surface energy correction using beta
        coefficients.
        
            E_surface = Σ_j β_lj / 4 Σ_i ρ_l[i]² Δx
        
        Parameters
        ----------
        x : numpy.ndarray, optional
            x-coordinates.  Default value is the stored x-coordinates.
        disregistry : numpy.ndarray, optional
            (N, 3) shaped array of disregistry vectors at each x-coordinate.
            Default value is the stored disregistry values.
        
        Returns
        -------
        float
            The surface energy for the dislocation.
        """
        # Default values are class properties
        if x is None:
            x = self.__x
        if disregistry is None:
            disregistry = self.__disregistry
            
        # Extract values
        d = disregistry
        dx = x[1] - x[0]
        beta = self.__beta
        cdiff = self.__cdiffsurface
        
        rho = self.disldensity(x=x, disregistry=d, cdiff=cdiff)[1]
        
        # Σ_j β_lj / 4 Σ_i ρ_l[i]² Δx
        return np.sum( np.inner(rho**2 * dx, beta) ) / 4
    
    def nonlocal_energy(self, x=None, disregistry=None):
        """
        Computes the nonlocal energy correction using alpha coefficient(s).
        
            E_nonlocal = Σ_m α_m Σ_i δ[i] (δ[i] - (δ[i+m] + δ[i-m]) / 2) Δx
        
        Parameters
        ----------
        x : numpy.ndarray, optional
            x-coordinates.  Default value is the stored x-coordinates.
        disregistry : numpy.ndarray, optional
            (N, 3) shaped array of disregistry vectors at each x-coordinate.
            Default value is the stored disregistry values.
        
        Returns
        -------
        float
            The nonlocal energy for the dislocation.
        """
        # Default values are class properties
        if x is None:
            x = self.__x
        if disregistry is None:
            disregistry = self.__disregistry
            
        # Extract values
        d = disregistry
        dx = x[1] - x[0]
        alpha = self.__alpha
        
        energy = 0.0
        
        # Σ_m α_m Σ_i δ[i] (δ[i] - (δ[i+m] + δ[i-m]) / 2) Δx
        for num, a in enumerate(alpha):
            m = num + 1
            dd = d[m:-m] - 0.5 * (d[2*m:] + d[:-2*m])
            energy += a * np.sum(d[m:-m] * dd * dx)
            
        return energy
    
    def total_energy(self, x=None, disregistry=None):
        """
        Computes the total energy for the dislocation.
        
            E_total = E_elastic + E_misfit + E_stress + E_surface + E_nonlocal
        
        Parameters
        ----------
        x : numpy.ndarray, optional
            x-coordinates.  Default value is the stored x-coordinates.
        disregistry : numpy.ndarray, optional
            (N, 3) shaped array of disregistry vectors at each x-coordinate.
            Default value is the stored disregistry values.
        
        Returns
        -------
        float
            The total energy for the dislocation.
        """
        # Default values are class properties
        if x is None:
            x = self.__x
        if disregistry is None:
            disregistry = self.__disregistry
            
        return (self.misfit_energy(x, disregistry)
                + self.elastic_energy(x, disregistry)
                + self.longrange_energy()
                + self.stress_energy(x, disregistry)
                + self.nonlocal_energy(x, disregistry)
                + self.surface_energy(x, disregistry))
    
    def load(self, model, gamma=None):
        """
        Load solution from a data model.
        
        Parameters
        ----------
        model : str or DataModelDict
            The semi-discrete-Peierls-Nabarro data model to load.
        gamma : atomman.defect.GammaSurface
            The gamma surface to use.
        """
        
        # Identify model element
        try:
            sdpn = DM(model).find('semidiscrete-variational-Peierls-Nabarro')
        except:
            sdpn = DM(model).find('semi-discrete-Peierls-Nabarro')
        # Load calculation parameters
        params = sdpn['parameter']
        self.__axes = uc.value_unit(params['axes'])
        self.__K_tensor = uc.value_unit(params['K_tensor'])
        self.__tau = uc.value_unit(params['tau'])
        self.__alpha = uc.value_unit(params['alpha'])
        if not isinstance(self.__alpha, list):
            self.__alpha = [self.__alpha]
        self.__beta = uc.value_unit(params['beta'])
        self.__cutofflongrange = uc.value_unit(params['cutofflongrange'])
        self.__burgers = uc.value_unit(params['burgers'])
        self.__fullstress = params['fullstress']
        self.__cdiffelastic = params['cdiffelastic']
        self.__cdiffsurface = params['cdiffsurface']
        self.__cdiffstress = params['cdiffstress']
        self.__min_method = params['min_method']
        self.__min_options = params['min_options']
        
        # Load gamma
        if gamma is None:
            try:
                gamma = GammaSurface(model)
            except:
                raise ValueError('No gamma surface in model or given!')
        elif not isinstance(gamma, GammaSurface):
            gamma = GammaSurface(gamma)
        self.__gamma = gamma
        
        # Load calculation solution
        solution = sdpn['solution']
        self.__x = uc.value_unit(solution['x'])
        self.__disregistry = uc.value_unit(solution['disregistry'])
    
    def model(self, length_unit='angstrom', energy_unit='eV',
              pressure_unit='GPa', include_gamma=False):
        """
        Generate a data model for the object.
        
        Parameters
        ----------
        length_unit : str
            The unit of length to save values as.  Default is 'angstrom'.
        energy_unit : str
            The unit of energy to save values as.  Default is 'eV'.
        pressure_unit : str
            The unit of pressure to save values as.  Default is 'GPa'.
        include_gamma : bool
            Flag indicating if the gamma surface data is to be included.
            Default value is False.
        """
        model = DM()
        model['semidiscrete-variational-Peierls-Nabarro'] = sdpn = DM()
        
        sdpn['parameter'] = params = DM()
        params['axes'] = uc.model(self.__axes, None)
        params['K_tensor'] = uc.model(self.__K_tensor, pressure_unit)
        params['tau'] = uc.model(self.__tau, pressure_unit)
        params['alpha'] = uc.model(self.__alpha, pressure_unit+'/'+length_unit)
        
        params['beta'] = uc.model(self.__beta, pressure_unit+'*'+length_unit)
        params['cdiffelastic'] = self.cdiffelastic
        params['cdiffsurface'] = self.cdiffsurface
        params['cdiffstress'] = self.cdiffstress
        params['cutofflongrange'] = uc.model(self.__cutofflongrange, length_unit)
        params['burgers'] = uc.model(self.__burgers, length_unit)
        params['fullstress'] = self.fullstress
        params['min_method'] = self.min_method
        params['min_options'] = self.min_options
        
        if include_gamma is True:
            sdpn['generalized-stacking-fault'] = self.__gamma.model(
                                                  length_unit=length_unit,
                                                  energy_unit=energy_unit,
                                                  pressure_unit=pressure_unit)
        
        sdpn['solution'] = solution = DM()
        solution['x'] = uc.model(self.__x, length_unit)
        solution['disregistry'] = uc.model(self.__disregistry, length_unit)
        
        return model


# coding: utf-8
"""
Attributes
----------
rootdir : str
    The absolute path to the iprPy package's root directory used to locate
    contained data files.
"""

# Standard Python libraries
import os

# Define rootdir
rootdir = os.path.dirname(os.path.abspath(__file__))

# Read version from VERSION file
with open(os.path.join(rootdir, 'VERSION')) as version_file:
    __version__ = version_file.read().strip()

# atomman imports
from . import unitconvert
from . import tools
from . import region
from . import lammps
from .dump import *
from .dump import __all__ as dump_all
from .core import *
from .core import __all__ as core_all
from .load import *
from .load import __all__ as load_all
from . import plot
from . import defect


__all__ = ['__version__', 'rootdir'] + dump_all + core_all + load_all
__all__.sort()

# Define default working units
unitconvert.reset_units(length = 'angstrom', mass = 'amu', energy='eV', charge='e')


# coding: utf-8
# Standard Python libraries
import datetime

# http://www.numpy.org/
import numpy as np

# https://pandas.pydata.org/
import pandas as pd

# https://github.com/usnistgov/DataModelDict
from DataModelDict import DataModelDict as DM

# atomman imports
from ..tools import uber_open_rmode

class Log(object):
    """Object for representing a LAMMPS log output"""
    
    def __init__(self, log_info=None):
        """
        Initializes a Log object.
        
        Parameters
        ----------
        log_info : str or file-like object, optional
            The LAMMPS log content to read in.  If None (default), then the
            Log object is created but empty.
        """
        
        # Initialize simulation properties
        self.__simulations = []
        self.__lammps_version = None
        self.__lammps_date = None
        
        # Read log data if supplied
        if log_info is not None:
            self.read(log_info)
    
    def read(self, log_info, append=True):
        """
        Parses a LAMMPS screen output/log file.
        
        Parameters
        ----------
        log_info : str or file-like object
            The LAMMPS log content to read in.
        append : bool, optional
            Flag indicating if the content being read in is appended to the
            current data (True, default), or if it overwrites any existing
            saved content (False).
        """
        
        # Reset properties and values if append is False
        if append is False:
            self.__simulations = []
            self.__lammps_version = None
            self.__lammps_date = None
        
        # Strings found directly before and after run and minimize simulations
        sim_start_trigger = ['Memory usage per processor =',
                             'Per MPI rank memory allocation (min/avg/max) =']
        sim_end_trigger = ['Loop time of']
        
        # Handle file names, strings and open file-like objects equivalently
        with uber_open_rmode(log_info) as log_info:
            
            # Initialize parameters
            headers = []
            footers = []
            i = 0
            
            # For all lines in file/output
            for line in log_info:
                line = line.decode('UTF-8')
                # Skip blank lines
                if len(line.split()) == 0:
                    continue
                
                # Save the LAMMPS version information
                if line[:8] == 'LAMMPS (' and self.lammps_version is None:
                    month = {'Jan': 1, 'Feb': 2, 'Mar': 3, 'Apr': 4, 
                             'May': 5, 'Jun': 6, 'Jul': 7, 'Aug': 8,
                             'Sep': 9, 'Oct': 10,'Nov': 11,'Dec': 12}
                    self.__lammps_version = line.strip()[8:-1]
                    d = self.lammps_version.split('-')[0].split()
                    self.__lammps_date = datetime.date(int(d[2]), month[d[1]], int(d[0]))
                
                # Check for strings listed prior to run and minimize simulations
                if any([trigger in line for trigger in sim_start_trigger]):
                    headers.append(i+1)
                
                # Check for strings listed after run and minimize simulations
                elif any([trigger in line for trigger in sim_end_trigger]):
                    footers.append(i-1)
                
                i += 1
            
            # Add last line to footers for incomplete logs
            footers.append(i)
            
            # Reset file pointer
            log_info.seek(0)
            
            # For all lines in file/output
            for header, footer in zip(headers, footers):
                
                # Initialize simulation data dictionary
                sim = {}
                
                # Use pandas to read all thermo data at once
                sim['thermo'] = pd.read_csv(log_info, header=header,
                                            nrows=footer-header,
                                            delim_whitespace=True,
                                            skip_blank_lines=True)
                
                # Reset file pointer
                log_info.seek(0)
                
                # Append simulation results
                self.__simulations.append(sim)
    
    @property
    def simulations(self):
        """list of dict: parsed data for each separate LAMMPS run/minimize action"""
        return self.__simulations
    
    @property
    def lammps_version(self):
        """str : The LAMMPS version used."""
        return self.__lammps_version
    
    @property
    def lammps_date(self):
        """datetime.date : The date associated with the LAMMPS version used."""
        return self.__lammps_date
    
    def flatten(self, style='last'):
        """
        Combines all simulations into one.
        
        Parameters
        ----------
        style : str, optional
            Specifies which values to use for duplicate time steps:
            - 'first' uses the values from the earliest simulation.
            - 'last' uses the values from the latest simulation (default).
        """
        # Check that all simulations with thermo data have step values
        for sim in self.simulations:
            if 'thermo' in sim and len(sim['thermo']) > 0:
                assert 'Step' in sim['thermo'], 'All simulation thermos must have Step key in order to flatten'
        
        # Combine the data into merged_df
        merged_df = self.simulations[0]['thermo']
        for i in range(1, len(self.simulations)):
            if 'thermo' in self.simulations[i]:
                thermo = self.simulations[i]['thermo']
                if style == 'first':
                    merged_df = pd.concat([merged_df, thermo[thermo.Step > merged_df.Step.max()]], ignore_index=True)
                elif style == 'last':
                    merged_df = pd.concat([merged_df[merged_df.Step < thermo.Step.min()], thermo], ignore_index=True)
                else:
                    raise ValueError('Unsupported style')
        
        self.__simulations = [{'thermo':merged_df}]
    
    def model(self, flatten=None):
        """
        Returns an XML/JSON equivalent data mode of the information
        
        Parameters
        ----------
        flatten : str or None, optional
            Specifies if the simulations are to be flattened, and which
            flatten style to use:
            - None does not flatten the simulations (default).
            - 'first' uses the values from the earliest simulation.
            - 'last' uses the values from the latest simulation.
            
        Returns
        -------
        DataModelDict
            The Log content in data model form.
        """
        if flatten is not None:
            self.flatten(flatten)
        
        # Create DataModelDict root
        log_model = DM()
        log_model['LAMMPS-log-thermo-data'] = DM()
        
        # Loop over all simulations
        for sim in self.simulations:
            sim_model = DM()
            
            # Convert to DataModelDict
            sim_model['thermo'] = DM()
            thermo = sim['thermo']
            for j in thermo:
                sim_model['thermo'][str(j)] = thermo[j].values.tolist()
            
            # Append simulation results to DataModelDict root
            log_model['LAMMPS-log-thermo-data'].append('simulation', sim_model)
            
        return log_model


# coding: utf-8
from . import style
from .Log import Log
from .NEBLog import NEBLog
from .Potential import Potential
from .run import run
from .normalize import normalize
from .checkversion import checkversion
__all__ = ['style', 'run', 'normalize', 'Potential', 'Log', 'NEBLog', 'checkversion']